@@ -24,7 +24,7 @@ package seccomp
 var defaultTemplate = []byte(`
 # Description: Allows access to app-specific directories and basic runtime
 #
-# The default seccomp policy is default deny with a whitelist of allowed
+# The default seccomp policy is default deny with an allowlist of allowed
 # syscalls. The default policy is intended to be safe for any application to
 # use and should be evaluated in conjunction with other security backends (eg
 # AppArmor). For example, a few particularly problematic syscalls that are left
@@ -200,8 +200,15 @@ inotify_rm_watch
 # input (man tty_ioctl), so we disallow it to prevent snaps plugging interfaces
 # with 'capability sys_admin' from interfering with other snaps or the
 # unconfined user's terminal.
+# similarly, TIOCLINUX allows to fake input as well (man ioctl_console) so
+# disallow that too
 # TODO: this should be scaled back even more
-ioctl - !TIOCSTI
+~ioctl - TIOCSTI
+~ioctl - TIOCLINUX
+# see CVE-2019-7303
+~ioctl - 4294967295|TIOCSTI
+~ioctl - 4294967295|TIOCLINUX
+ioctl
 
 io_cancel
 io_destroy
@@ -217,6 +224,10 @@ ioprio_get
 
 ipc
 kill
+# kcmp is guarded in the kernel via ptrace with PTRACE_MODE_READ_REALCREDS
+# such that the calling process must already be able to ptrace the target
+# processes and so this is safe.
+kcmp - - KCMP_FILE
 link
 linkat
 
@@ -444,6 +455,7 @@ socket AF_LOCAL
 socket AF_INET
 socket AF_INET6
 socket AF_IPX
+socket AF_XDP
 socket AF_X25
 socket AF_AX25
 socket AF_ATMPVC
@@ -695,14 +707,19 @@ setresuid32 u:root u:root -1
 // Template for privilege drop and chown operations. This intentionally does
 // not support all combinations of users or obscure combinations (we can add
 // combinations as users dictate). Eg, these are supported:
-//   chown foo:foo
-//   chown foo
-//   chgrp foo
+//
+//	chown foo:foo
+//	chown foo
+//	chgrp foo
+//
 // but these are not:
-//   chown foo:bar
-//   chown bar:foo
+//
+//	chown foo:bar
+//	chown bar:foo
+//
 // For now, users who want 'foo:bar' can do:
-//   chown foo ; chgrp bar
+//
+//	chown foo ; chgrp bar
 var privDropAndChownSyscalls = `
 # allow setgid to ###GROUP###
 setgid g:###GROUP###
