@@ -98,7 +98,7 @@ func (t *target) setups(st *state.State, opts Options) (SnapSetup, []ComponentSe
 			CompPath:     comp.CompPath,
 			DownloadInfo: comp.DownloadInfo,
 
-			componentInstallFlags: componentInstallFlags{
+			ComponentInstallFlags: ComponentInstallFlags{
 				// if we're removing the snap, then we should remove the
 				// components too
 				RemoveComponentPath:   opts.Flags.RemoveSnapPath,
@@ -155,6 +155,11 @@ func (t *target) setups(st *state.State, opts Options) (SnapSetup, []ComponentSe
 		InstanceKey:        t.info.InstanceKey,
 		ExpectedProvenance: t.info.SnapProvenance,
 		Registries:         registries,
+		auxStoreInfo: auxStoreInfo{
+			Media: t.info.Media,
+			// XXX we store this for the benefit of old snapd
+			Website: t.info.Website(),
+		},
 	}, compsups, nil
 }
 
@@ -372,8 +377,6 @@ func componentTargetsFromActionResult(action string, sar store.SnapActionResult,
 	for _, comp := range requested {
 		res, ok := mapping[comp]
 		if !ok {
-			// TODO:COMPS: make sure this branch is tested when we add support for
-			// losing components during a refresh
 			// during a refresh, we will not install components that don't exist
 			// in the new revision
 			if action == "refresh" {
@@ -586,6 +589,14 @@ func InstallOne(ctx context.Context, st *state.State, goal InstallGoal, opts Opt
 	return infos[0], tasksets[0], nil
 }
 
+func sortComponentsOnTargets(targets []target) {
+	for _, t := range targets {
+		sort.Slice(t.components, func(i, j int) bool {
+			return t.components[i].ComponentName() < t.components[j].ComponentName()
+		})
+	}
+}
+
 // InstallWithGoal installs the snap/set of snaps specified by the given
 // InstallGoal.
 //
@@ -624,13 +635,7 @@ func InstallWithGoal(ctx context.Context, st *state.State, goal InstallGoal, opt
 		return nil, nil, ErrExpectedOneSnap
 	}
 
-	for _, t := range targets {
-		// sort the components by name to ensure we always install components in the
-		// same order
-		sort.Slice(t.components, func(i, j int) bool {
-			return t.components[i].ComponentName() < t.components[j].ComponentName()
-		})
-	}
+	sortComponentsOnTargets(targets)
 
 	installInfos := make([]minimalInstallInfo, 0, len(targets))
 	for _, t := range targets {
@@ -857,7 +862,11 @@ func (p *updatePlan) revisionChanges(st *state.State, opts Options) ([]*snap.Inf
 
 	changes := make([]*snap.Info, 0, len(updates))
 	for _, up := range updates {
-		if up.revisionSatisfied() {
+		ok, err := up.revisionSatisfied()
+		if err != nil {
+			return nil, err
+		}
+		if ok {
 			continue
 		}
 
@@ -1000,6 +1009,8 @@ func UpdateWithGoal(ctx context.Context, st *state.State, goal UpdateGoal, filte
 		return nil, nil, err
 	}
 
+	sortComponentsOnTargets(plan.targets)
+
 	if opts.ExpectOneSnap && len(plan.targets) != 1 {
 		return nil, nil, ErrExpectedOneSnap
 	}
@@ -1189,7 +1200,7 @@ type PathSnap struct {
 	// SideInfo contains extra information about the snap.
 	SideInfo *snap.SideInfo
 	// Components is a mapping of component side infos to paths that should be
-	// intsalled alongside this snap.
+	// installed alongside this snap.
 	Components map[*snap.ComponentSideInfo]string
 }
 
@@ -1236,10 +1247,6 @@ func (p *pathUpdateGoal) toUpdate(_ context.Context, st *state.State, opts Optio
 			return updatePlan{}, err
 		}
 
-		// TODO:COMPS: remove this once we are ready to handle components during
-		// refresh
-		t.components = nil
-
 		targets = append(targets, t)
 		names = append(names, sn.InstanceName)
 	}
@@ -1275,7 +1282,7 @@ func targetForPathSnap(update PathSnap, snapst SnapState, opts Options) (target,
 		return target{}, fmt.Errorf("cannot install local snap %q: %v != %v (revision mismatch)", update.InstanceName, update.RevOpts.Revision, si.Revision)
 	}
 
-	if update.RevOpts.Channel != "" && update.SideInfo.Channel != "" && update.RevOpts.Channel != update.SideInfo.Channel {
+	if update.RevOpts.Channel != "" && si.Channel != "" && update.RevOpts.Channel != si.Channel {
 		return target{}, fmt.Errorf("cannot install local snap %q: %v != %v (channel mismatch)", update.InstanceName, update.RevOpts.Channel, si.Channel)
 	}
 
@@ -1293,8 +1300,7 @@ func targetForPathSnap(update PathSnap, snapst SnapState, opts Options) (target,
 		update.RevOpts.Channel = update.SideInfo.Channel
 	}
 
-	channel, err := resolveChannel(update.InstanceName, trackingChannel, update.RevOpts.Channel, opts.DeviceCtx)
-	if err != nil {
+	if err := update.RevOpts.resolveChannel(update.InstanceName, trackingChannel, opts.DeviceCtx); err != nil {
 		return target{}, err
 	}
 
@@ -1306,7 +1312,7 @@ func targetForPathSnap(update PathSnap, snapst SnapState, opts Options) (target,
 	return target{
 		setup: SnapSetup{
 			SnapPath:  update.Path,
-			Channel:   channel,
+			Channel:   update.RevOpts.Channel,
 			CohortKey: update.RevOpts.CohortKey,
 		},
 		info:       info,
