@@ -5,9 +5,6 @@ set -x
 # handle errors in general.
 set -e
 
-# shellcheck source=tests/lib/quiet.sh
-. "$TESTSLIB/quiet.sh"
-
 # shellcheck source=tests/lib/pkgdb.sh
 . "$TESTSLIB/pkgdb.sh"
 
@@ -62,32 +59,49 @@ create_test_user(){
     echo >> /etc/sudoers
     echo 'test ALL=(ALL) NOPASSWD:ALL' >> /etc/sudoers
 
-    chown test.test -R "$SPREAD_PATH"
-    chown test.test "$SPREAD_PATH/../"
+    chown test:test -R "$SPREAD_PATH"
+    chown test:test "$SPREAD_PATH/../"
 }
 
 build_deb(){
+    newver="$(dpkg-parsechangelog --show-field Version)"
+
+    case "$SPREAD_SYSTEM" in
+        ubuntu-fips-*)
+            newver="${newver}+fips"
+            FIPS_BUILD_OPTION=fips
+            ;;
+    esac
     # Use fake version to ensure we are always bigger than anything else
-    dch --newversion "1337.$(dpkg-parsechangelog --show-field Version)" "testing build"
+    dch --newversion "1337.$newver" "testing build"
 
     if os.query is-debian sid; then
         # ensure we really build without vendored packages
-        rm -rf vendor/*/*
+        mv ./vendor /tmp
     fi
 
     unshare -n -- \
-            su -l -c "cd $PWD && DEB_BUILD_OPTIONS='nocheck testkeys' dpkg-buildpackage -tc -b -Zgzip" test
+            su -l -c "cd $PWD && DEB_BUILD_OPTIONS='nocheck testkeys ${FIPS_BUILD_OPTION}' dpkg-buildpackage -tc -b -Zgzip -uc -us" test
     # put our debs to a safe place
     cp ../*.deb "$GOHOME"
+
+    if os.query is-debian sid; then
+        # restore vendor dir, it's needed by e.g. fakestore
+        mv /tmp/vendor ./
+    fi
 }
 
 build_rpm() {
     distro=$(echo "$SPREAD_SYSTEM" | awk '{split($0,a,"-");print a[1]}')
     release=$(echo "$SPREAD_SYSTEM" | awk '{split($0,a,"-");print a[2]}')
-    if os.query is-amazon-linux; then
+    if os.query is-amazon-linux 2; then
         distro=amzn
         release=2
     fi
+    if os.query is-amazon-linux 2023; then
+        distro=amzn
+        release=2023
+    fi
     arch=x86_64
     base_version="$(head -1 debian/changelog | awk -F '[()]' '{print $2}')"
     version="1337.$base_version"
@@ -95,12 +109,12 @@ build_rpm() {
     rpm_dir=$(rpm --eval "%_topdir")
     pack_args=
     case "$SPREAD_SYSTEM" in
-        fedora-*|amazon-*|centos-*)
-            ;;
         opensuse-*)
             # use bundled snapd*.vendor.tar.xz archive
             pack_args=-s
             ;;
+        fedora-*|amazon-*|centos-*)
+            ;;
         *)
             echo "ERROR: RPM build for system $SPREAD_SYSTEM is not yet supported"
             exit 1
@@ -116,22 +130,18 @@ build_rpm() {
 
     # Cleanup all artifacts from previous builds
     rm -rf "$rpm_dir"/BUILD/*
+    # Install build dependencies
+    distro_install_package rpmdevtools
+    # XXX we should pass --with testkeys for completeness, but older versions of
+    # rpmspec do not support it, and in any case testkeys does not result in any
+    # additional build packages
+    # shellcheck disable=SC2046
+    distro_install_package $(rpmspec -q --buildrequires "$packaging_path/snapd.spec")
 
     # Build our source package
     unshare -n -- \
             rpmbuild --with testkeys -bs "$rpm_dir/SOURCES/snapd.spec"
 
-    # .. and we need all necessary build dependencies available
-    deps=()
-    IFS=$'\n'
-    for dep in $(rpm -qpR "$rpm_dir"/SRPMS/snapd-1337.*.src.rpm); do
-      if [[ "$dep" = rpmlib* ]]; then
-         continue
-      fi
-      deps+=("$dep")
-    done
-    distro_install_package "${deps[@]}"
-
     # And now build our binary package
     unshare -n -- \
             rpmbuild \
@@ -203,6 +213,11 @@ download_from_published(){
     done
 }
 
+download_from_gce_bucket(){
+    curl -o "${SPREAD_SYSTEM}.tar" "https://storage.googleapis.com/snapd-spread-tests/snapd-tests/packages/${SPREAD_SYSTEM}.tar"
+    tar -xf "${SPREAD_SYSTEM}.tar" -C "$PROJECT_PATH"/..
+}
+
 install_dependencies_from_published(){
     local published_version="$1"
 
@@ -211,6 +226,36 @@ install_dependencies_from_published(){
     done
 }
 
+install_snapd_rpm_dependencies(){
+    SRC_PATH=$1
+    deps=()
+    IFS=$'\n'
+    for dep in $(rpm -qpR "$SRC_PATH"); do
+        if [[ "$dep" = rpmlib* ]]; then
+            continue
+        fi
+        deps+=("$dep")
+    done
+    distro_install_package "${deps[@]}"
+}
+
+install_dependencies_gce_bucket(){
+    case "$SPREAD_SYSTEM" in
+        ubuntu-*|debian-*)
+            cp "$PROJECT_PATH"/../*.deb "$GOHOME"
+            ;;
+        fedora-*|opensuse-*|amazon-*|centos-*)
+            install_snapd_rpm_dependencies "$PROJECT_PATH"/../snapd-1337.*.src.rpm
+            # sources are not needed to run the tests
+            rm "$PROJECT_PATH"/../snapd-1337.*.src.rpm
+            find "$PROJECT_PATH"/.. -name '*.rpm' -exec cp -v {} "${GOPATH%%:*}" \;
+            ;;
+        arch-*)
+            cp "$PROJECT_PATH"/../snapd*.pkg.tar.* "${GOPATH%%:*}"
+            ;;
+    esac
+}
+
 ###
 ### Prepare / restore functions for {project,suite}
 ###
@@ -232,6 +277,8 @@ prepare_project() {
     # no need to modify anything further for autopkgtest
     # we want to run as pristine as possible
     if [ "$SPREAD_BACKEND" = autopkgtest ]; then
+        create_test_user
+        systemctl enable --now snapd.socket
         exit 0
     fi
 
@@ -250,8 +297,20 @@ prepare_project() {
 
     # declare the "quiet" wrapper
 
-    if [ "$SPREAD_BACKEND" = external ]; then
-        chown test.test -R "$PROJECT_PATH"
+    if [ "$SPREAD_BACKEND" = "external" ]; then
+        chown test:test -R "$PROJECT_PATH"
+        exit 0
+    fi
+
+    if [ "$SPREAD_BACKEND" = "testflinger" ]; then
+        if os.query is-core-ge 24; then
+            useradd --uid 12345 --create-home --extrausers test
+        else
+            adduser --uid 12345 --extrausers --quiet --disabled-password --gecos '' test
+        fi
+        echo test:ubuntu | sudo chpasswd
+        echo 'test ALL=(ALL) NOPASSWD:ALL' | sudo tee /etc/sudoers.d/create-user-test
+        chown test:test -R "$PROJECT_PATH"
         exit 0
     fi
 
@@ -278,6 +337,9 @@ prepare_project() {
     create_test_user
 
     distro_update_package_db
+    # XXX this should be part of the image update in spread-images
+    # remove any packages that are marked for auto removal before running any tests
+    distro_auto_remove_packages
 
     if os.query is-arch-linux; then
         # perform system upgrade on Arch so that we run with most recent kernel
@@ -331,10 +393,10 @@ prepare_project() {
         rm -rf vendor/*/
 
         # and create a fake upstream tarball
-        tar -c -z -f ../snapd_"$(dpkg-parsechangelog --show-field Version|cut -d- -f1)".orig.tar.gz --exclude=./debian --exclude=./.git .
+        tar -c -z -f ../snapd_"$(dpkg-parsechangelog --show-field Version|cut -d- -f1)".orig.tar.gz --exclude=./debian --exclude=./.git --exclude='*.pyc' .
 
         # and build a source package - this will be used during the sbuild test
-        dpkg-buildpackage -S --no-sign
+        dpkg-buildpackage -S -uc -us
     fi
 
     # so is ubuntu-14.04
@@ -361,6 +423,14 @@ prepare_project() {
         quiet eatmydata apt-get install -y --force-yes apparmor libapparmor1 seccomp libseccomp2 systemd cgroup-lite util-linux
     fi
 
+    # ubuntu-16.04 is EOL so the updated go-1.18 is only available via
+    # the ppa:snappy-dev/image ppa for now. if needed the package could
+    # be copied from the PPA to the ESM archive.
+    if os.query is-xenial; then
+        quiet add-apt-repository ppa:snappy-dev/image
+        quiet eatmydata apt-get update
+    fi
+
     # WORKAROUND for older postrm scripts that did not do
     # "rm -rf /var/cache/snapd"
     rm -rf /var/cache/snapd/aux
@@ -423,14 +493,16 @@ prepare_project() {
     esac
 
     restart_logind=
-    if [ "$(systemctl --version | awk '/systemd [0-9]+/ { print $2 }')" -lt 246 ]; then
+    local systemd_ver
+    systemd_ver="$(systemctl --version | awk '/systemd [0-9]+/ { print $2 }' | cut -f1 -d"~")"
+    if [ "$systemd_ver" -lt 246 ]; then
         restart_logind=maybe
     fi
 
     install_pkg_dependencies
 
     if [ "$restart_logind" = maybe ]; then
-        if [ "$(systemctl --version | awk '/systemd [0-9]+/ { print $2 }')" -ge 246 ]; then
+        if [ "$systemd_ver" -ge 246 ]; then
             restart_logind=yes
         else
             restart_logind=
@@ -483,50 +555,39 @@ prepare_project() {
     # base on the packaging. In Fedora/Suse this is handled via mock/osc
     case "$SPREAD_SYSTEM" in
         debian-*|ubuntu-*)
-            best_golang=golang-1.13
-            if [[ "$SPREAD_SYSTEM" == debian-9-* ]]; then
-                echo "debian-9 tests disabled (no golang-1.13)"
-                exit 1
-            elif [[ "$SPREAD_SYSTEM" == debian-10-* ]]; then
-                # debian-10 needs backports for golang-1.13
-                echo "deb http://deb.debian.org/debian buster-backports main" >> /etc/apt/sources.list
-                apt update
-                # dh-golang must come from backports, gdebi/apt cannot
-                # resolve this on their own
-                apt install -y -t buster-backports dh-golang
-                # we need the specific golang-1.13 here, not golang-go
-                sed -i -e "s/golang-go (>=2:1.13).*,/${best_golang},/" ./debian/control
-            fi
+            best_golang=golang-1.18
+            case "$SPREAD_SYSTEM" in
+                ubuntu-fips-*)
+                    # we are limited by the FIPS variants of go toolchain
+                    # available from the PPA, and we need to match the Go
+                    # version expected by during FIPS build of the deb, which
+                    # currently expects 1.21, see:
+                    # https://launchpad.net/~ubuntu-toolchain-r/+archive/ubuntu/golang-fips
+                    best_golang=golang-1.21
+                    quiet apt install -y golang-1.21
+                    ;;
+            esac
             # in 16.04: "apt build-dep -y ./" would also work but not on 14.04
-            gdebi --quiet --apt-line ./debian/control | quiet xargs -r eatmydata apt-get install -y
-            # The go 1.13 backport is not using alternatives or anything else so
+            gdebi --quiet --apt-line ./debian/control >deps.txt
+            quiet xargs -r eatmydata apt-get install -y < deps.txt
+            # The go 1.18 backport is not using alternatives or anything else so
             # we need to get it on path somehow. This is not perfect but simple.
             if [ -z "$(command -v go)" ]; then
-                # the path filesystem path is: /usr/lib/go-1.13/bin
+                # the path filesystem path is: /usr/lib/go-1.18/bin
                 ln -s "/usr/lib/${best_golang/lang/}/bin/go" /usr/bin/go
             fi
             ;;
     esac
 
     # Retry go mod vendor to minimize the number of connection errors during the sync
-    for _ in $(seq 10); do
-        if quiet go mod vendor; then
-            break
-        fi
-        sleep 1
-    done
+    retry -n 10 go mod vendor
     # Update C dependencies
-    for _ in $(seq 10); do
-        if (cd c-vendor && ./vendor.sh); then
-            break
-        fi
-        sleep 1
-    done
+    ( cd c-vendor && retry -n 10 ./vendor.sh )
 
     # go mod runs as root and will leave strange permissions
-    chown test.test -R "$SPREAD_PATH"
+    chown test:test -R "$SPREAD_PATH"
 
-    if [ -z "$SNAPD_PUBLISHED_VERSION" ]; then
+    if [ "$BUILD_SNAPD_FROM_CURRENT" = true ]; then
         case "$SPREAD_SYSTEM" in
             ubuntu-*|debian-*)
                 build_deb
@@ -542,9 +603,12 @@ prepare_project() {
                 exit 1
                 ;;
         esac
-    else
+    elif [ -n "$SNAPD_PUBLISHED_VERSION" ]; then
         download_from_published "$SNAPD_PUBLISHED_VERSION"
         install_dependencies_from_published "$SNAPD_PUBLISHED_VERSION"
+    else
+        download_from_gce_bucket
+        install_dependencies_gce_bucket
     fi
 
     # Build fakestore.
@@ -586,7 +650,9 @@ prepare_project_each() {
 prepare_suite() {
     # shellcheck source=tests/lib/prepare.sh
     . "$TESTSLIB"/prepare.sh
-    if os.query is-core; then
+    # os.query cannot be used because first time the suite is prepared, the current system
+    # is classic ubuntu, so it is needed to check the system set in $SPREAD_SYSTEM
+    if is_test_target_core; then
         prepare_ubuntu_core
     else
         prepare_classic
@@ -614,6 +680,16 @@ prepare_suite_each() {
 
     # Save all the installed packages
     if os.query is-classic; then
+        # lxd-installer is in cloud images starting from 24.04. This package
+        # installs lxd when any lxc command is run. This caused problems
+        # because if we install snapcraft & lxd, in the restore step lxd is
+        # removed, and after that snapcraft is removed. However, snapcraft's
+        # remove hook calls lxd and triggers a new installation of lxd, and in
+        # turn when we try to remove core22, it fails as lxd has been
+        # re-installed and depends on that base. Therefore, we remove it to
+        # prevent these issues, and we do that before we get the list of
+        # installed packages to make sure we do not re-install it again.
+        apt remove -y --purge lxd-installer || true
         tests.pkgs list-installed > installed-initial.pkgs
     fi
 
@@ -621,7 +697,7 @@ prepare_suite_each() {
     tests.backup prepare
 
     # save the job which is going to be executed in the system
-    echo -n "$SPREAD_JOB " >> "$RUNTIME_STATE_PATH/runs"
+    echo -n "${SPREAD_JOB:-} " >> "$RUNTIME_STATE_PATH/runs"
 
     # Restart journal log and reset systemd journal cursor.
     systemctl reset-failed systemd-journald.service
@@ -641,12 +717,11 @@ prepare_suite_each() {
     fi
 
     if [[ "$variant" = full ]]; then
-        # shellcheck source=tests/lib/prepare.sh
-        . "$TESTSLIB"/prepare.sh
         if os.query is-classic; then
+            # shellcheck source=tests/lib/prepare.sh
+            . "$TESTSLIB"/prepare.sh
             prepare_each_classic
         fi
-        prepare_memory_limit_override
     fi
 
     case "$SPREAD_SYSTEM" in
@@ -713,9 +788,13 @@ restore_suite_each() {
     done
 
     if [[ "$variant" = full ]]; then
-        # reset the failed status of snapd, snapd.socket, and snapd.failure.socket
-        # to prevent hitting the system restart rate-limit for these services
-        systemctl reset-failed snapd.service snapd.socket snapd.failure.service
+        # Reset the failed status of snapd, snapd.socket, and snapd.failure.socket
+        # to prevent hitting the system restart rate-limit for these services.
+        systemctl reset-failed snapd.service snapd.socket
+        # This unit may not be present, it is masked on some systems.
+        if systemctl status snapd.failure.service; then
+            systemctl reset-failed snapd.failure.service
+        fi
     fi
 
     if [[ "$variant" = full ]]; then
@@ -735,9 +814,7 @@ restore_suite_each() {
 restore_suite() {
     # shellcheck source=tests/lib/reset.sh
     if [ "$REMOTE_STORE" = staging ]; then
-        # shellcheck source=tests/lib/store.sh
-        . "$TESTSLIB"/store.sh
-        teardown_staging_store
+        "$TESTSTOOLS"/store-state teardown-staging-store
     fi
 
     if os.query is-classic; then
